# coding=utf-8

"""Command line processing"""


import argparse
from sksurgeryspeech import __version__
from sksurgeryspeech.ui import sksurgeryspeech_demo


def main(args=None):
    """Entry point for scikit-surgeryspeech application"""

    parser = argparse.ArgumentParser(description='scikit-surgeryspeech')

    version_string = __version__
    friendly_version_string = version_string if version_string else 'unknown'
    parser.add_argument(
        "--version",
        action='version',
        version='scikit-surgeryspeech version ' + friendly_version_string)

    parser.add_argument(
        "-c", "--config",
        required=True,
        type=str,
        help="Configuration file")

    args = parser.parse_args(args)

    demo = sksurgeryspeech_demo.SpeechRecognitionDemo(args.config)
    demo.run_demo()



"""
Speech API algorithm
"""
# pylint: disable=no-name-in-module,import-error

import logging
import json
import struct
from datetime import datetime
import pyaudio
import pvporcupine
import speech_recognition as sr
from PySide2.QtCore import QObject, Signal, Slot, QThread, QTimer


LOGGER = logging.getLogger("voice_recognition_logger")


class VoiceRecognitionService(QObject):
    """
    Voice Recognition service which takes an microphone input and converts it
    to text by using the Google Cloud Speech-to-Text API.

    Configuration dictionary must contain the following keys:

    porcupine dynamic library path: \
        Porcupine/lib/<operating_system>/<processor_type>/<library_file>

    porcupine model file path: \
        Porcupine/lib/common/porcupine_params.pv

    porcupine keyword file(s): \
        Porcupine/resources/keyword_files/<operating_system>/<keyword>

    optional keys:

    google credentials file: json file with google cloud api credentials

    recogniser: api to use, options are sphinx, google, google_cloud, \
        bing, houdify, ibm, wit

    sphinx keywords: a list of keywords and sensitivities for sphinx
    timeout for command: default None
    """

    start_listen = Signal()
    stop_timer = Signal()
    google_api_not_understand = Signal()
    google_api_request_failure = Signal(str)
    voice_command = Signal(str)
    start_processing_request = Signal()

    def __init__(self, config: dict):
        """
        Construct the service.

        """
        LOGGER.info("Creating Voice Recognition Service")
        # Need this for SignalInstance
        super().__init__()

        self.timeout_for_command = config.get("timeout for command", None)

        library_path = config.get("porcupine dynamic library path", None)
        if library_path is None:
            raise KeyError("Config must contain porcupine dynamic",
                           " library path")

        model_file_path = config.get("porcupine model file path", None)
        if model_file_path is None:
            raise KeyError("Config must contain porcupine model file path")

        keyword_file_paths = config.get("porcupine keyword file", None)
        if keyword_file_paths is None:
            raise KeyError("Config must contain porcupine keyword file")

        self.recogniser = config.get("recogniser", "sphinx")
        self.sphinx_keywords = config.get("sphinx keywords", None)

        self.recognizer = sr.Recognizer()

        sensitivities = config.get("sensitivities", [1.0])
        self.interval = config.get("interval", 10)

        self.handle = pvporcupine.create(
            library_path=library_path,
            model_path=model_file_path,
            keyword_paths=keyword_file_paths,
            sensitivities=sensitivities)

        audio = pyaudio.PyAudio()
        self.audio_stream = \
            audio.open(rate=self.handle.sample_rate,
                       channels=1,
                       format=pyaudio.paInt16,
                       input=True,
                       frames_per_buffer=self.handle.frame_length)

        #  this is to add the credentials for the google cloud api
        #  set the environment variable GOOGLE_APPLICATION_CREDENTIALS
        #  to the path  of your json file with credentials
        key_file_path = config.get('google credentials file', None)
        self.credentials = None
        if key_file_path is not None:
            with open(key_file_path, 'r', encoding='utf-8') as file:
                self.credentials = file.read()

                #r aises a ValueError if the credential file isn't a valid json
                json.loads(self.credentials)

        # Creating timer later, in the context of the running thread.
        self.timer = None

        LOGGER.info("Created Voice Recognition Service")

    def run(self):
        """
        Entry point for the QThread which starts the timer to listen in the
        background
        """
        LOGGER.info("run method executed")

        # Creating the timer in the context of the running thread.
        self.timer = QTimer()
        self.timer.setInterval(self.interval)
        self.timer.timeout.connect(self.listen_for_keyword)
        self.stop_timer.connect(self.__stop)

        #  start the timer to start the background listening
        self.timer.start()

    def request_stop(self):
        """
        Called by external client to stop timer.
        """
        LOGGER.info("Requesting VoiceRecognitionService to stop timer.")
        self.stop_timer.emit()
        QThread.msleep(self.interval * 3)
        while self.timer.isActive():
            QThread.msleep(self.interval * 3)
        LOGGER.info("Requested VoiceRecognitionService to stop timer.")

    @Slot()
    def __stop(self):
        LOGGER.info("Stopping VoiceRecognitionService timer.")
        self.timer.stop()
        QThread.msleep(self.interval * 3)
        LOGGER.info("Stopped VoiceRecognitionService timer.")

    def listen_for_keyword(self):
        """
        This method is called every 100 milliseconds by the QThread running and
        listens for the keyword
        """
        pcm = self.audio_stream.read(self.handle.frame_length)
        pcm = struct.unpack_from("h" * self.handle.frame_length, pcm)
        result = self.handle.process(pcm)
        if result >= 0:
            #  when the keyword gets detected, the user can input a command
            LOGGER.info('[%s] detected keyword', str(datetime.now()))
            self.start_listen.emit()
            self.listen_to_command()

    def listen_to_command(self):
        """
        This method gets called when a specific command is said.
        It then listens for specific commands and converts them to QT Signals
        """
        #  listen to a single command
        with sr.Microphone() as source:
            audio = self.recognizer \
                .listen(source, phrase_time_limit=self.timeout_for_command)
        try:
            #  convert command to string,
            #  this string should later be used to fire a certain GUI command
            self.start_processing_request.emit()
            words = self._recognise(audio)

            self.voice_command.emit(words)
        except sr.UnknownValueError:
            self.google_api_not_understand.emit()
        except sr.RequestError as exception:
            self.google_api_request_failure.emit(str(exception))

    def _recognise(self, audio):
        words = ""
        if self.recogniser == "sphinx":
            words = self.recognizer.recognize_sphinx(
                audio, keyword_entries=self.sphinx_keywords)
        elif self.recogniser == "google_cloud":
            words = self.recognizer.recognize_google_cloud(
                audio, credentials_json=self.credentials)
        elif self.recogniser == "google":
            words = self.recognizer.recognize_google(audio)
        elif self.recogniser == "bing":
            raise NotImplementedError(
                "Key credentials for bing not set up")
            #something like this, but might need to change credentials
            #words = self.recognizer.recognize_bing(audio, key=self.credentials)
        elif self.recogniser == "houndify":
            raise NotImplementedError(
                "Key credentials for houndify not set up")
            #something like this, but might need to change credentials
            #words = self.recognizer.recognize_houndify(
            #    audio, client_id=self.credentials,
            #    client_key = self.credentials)
        elif self.recogniser == "ibm":
            raise NotImplementedError(
                "Key credentials for ibm not set up")
            #something like this, but might need to change credentials
            #words = recognizer.recognize_ibm(
            #    audio, username=notset, password=notset)
        elif self.recogniser == "wit":
            raise NotImplementedError(
                "Key credentials for wit not set up")
            #something like this, but might need to change credentials
            #words = self.recognizer.recognize_wit(audio, key=self.credentials)
        else:
            raise ValueError("Unrecognised recogniser", self.recogniser)

        return words



# coding=utf-8
"""
Demo for the Speech API module
"""
# pylint: disable=c-extension-no-member, no-name-in-module, no-self-use
import sys
import logging
import PySide2.QtCore
from sksurgerycore.configuration.configuration_manager import \
        ConfigurationManager
from sksurgeryspeech.algorithms import voice_recognition_service as speech_api

LOGGER = logging.getLogger("voice_recognition_logger")


class VoiceListener(PySide2.QtCore.QObject):

    """
    Class which contains the slots for the demo application
    """

    @PySide2.QtCore.Slot()
    def on_start_listen(self):
        """
        Slot when the user says the keyword
        """
        LOGGER.info("Listening for command")


    @PySide2.QtCore.Slot()
    def on_google_api_not_understand(self):
        """
        Slot if the google api doesn't understand audio
        """
        LOGGER.info("Google Speech Recognition could not understand audio")

    @PySide2.QtCore.Slot()
    def on_google_api_request_failure(self, exception):
        """
        Slot if something with the google api went wrong
        """
        LOGGER.info("Could not request results from Google Speech "
                    "Recognition service; %s", exception)

    @PySide2.QtCore.Slot()
    def on_start_processing_request(self):
        """
        Slot when the request is sent to Google API
        """
        LOGGER.info("Processing...")


class SpeechRecognitionDemo(PySide2.QtCore.QObject):

    """
    Demo class for the Speech API module
    """
    def __init__(self, config_file):
        """
        Constructor.
        """
        super().__init__()

        #  set up the logger
        voice_recognition_logger = logging.getLogger("voice_recognition_logger")
        voice_recognition_logger.setLevel(logging.INFO)
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)
        file_handler = logging.FileHandler('voice_recognition_log.log')
        file_handler.setLevel(logging.INFO)
        voice_recognition_logger.addHandler(console_handler)
        voice_recognition_logger.addHandler(file_handler)

        #  create VoiceRecognitionService()

        configurer = ConfigurationManager(config_file)
        config = configurer.get_copy()
        self.voice_recognition = speech_api.VoiceRecognitionService(config)
        #  create VoiceListener() which in this example has all the slots to
        #  react to the signals from the VoiceRecognitionService()
        self.listener = VoiceListener()

        #  Move the VoiceRecognitionService() to a separate thread so it doesn't
        #  block the main thread
        self.listener_thread = PySide2.QtCore.QThread(self)
        self.voice_recognition.moveToThread(self.listener_thread)
        self.listener_thread.started.connect(self.voice_recognition.run)

        #  connect the Signals emitted by the VoiceRecognitionService()
        #  with the Slots of the VoiceListener
        self.voice_recognition.start_listen\
            .connect(self.listener.on_start_listen)
        self.voice_recognition.google_api_not_understand\
            .connect(self.listener.on_google_api_not_understand)
        self.voice_recognition.google_api_request_failure\
            .connect(self.listener.on_google_api_request_failure)
        self.voice_recognition.start_processing_request\
            .connect(self.listener.on_start_processing_request)
        #connect this to our own slot
        self.voice_recognition.voice_command.connect(self.on_voice_signal)
    def run_demo(self):
        """
        Entry point to run the demo
        """
        #  instantiate the QCoreApplication
        app = PySide2.QtCore.QCoreApplication()
        #  this is the main call to start the background thread listening
        self.listener_thread.start()
        #  start the application, meaning starting the infinite Event Loop which
        #  stops when the user says "start" followed by "quit"

        return sys.exit(app.exec_())

    @PySide2.QtCore.Slot()
    def on_voice_signal(self, input_string):
        """
        Slot for the quit signal
        Quits application
        """
        LOGGER.info("Got voice signal, %s", input_string)
        if "exit" in input_string or "quit" in input_string:

            LOGGER.info("Quit signal caught... Exit application")
            self.voice_recognition.request_stop()
            self.listener_thread.quit()
            while not self.listener_thread.isFinished():
                PySide2.QtCore.QThread.msleep(100 * 3)
                LOGGER.info("Waiting for listener thread to stop")
            PySide2.QtCore.QCoreApplication.quit()



# coding=utf-8
"""
Setup for scikit-surgeryspeech
"""

from setuptools import setup, find_packages
import versioneer
import os

# Get the long description
with open('README.rst') as f:
    long_description = f.read()

# Read the docs can't install pyaudio
in_rtd_build = os.getenv('IN_READTHEDOCS')
if in_rtd_build == 'true':
    install_requires = [
                'pyaudio',
                'SpeechRecognition',
                'google-api-python-client',
                'oauth2client',
                'PySide2',
                'scikit-surgerycore',
                'pvporcupine<2.0.0'
            ]
else:
    install_requires = [
            'pyaudio',
            'SpeechRecognition',
            'google-api-python-client',
            'oauth2client',
            'PySide2',
            'scikit-surgerycore',
            'pvporcupine'
        ]

    setup(
        name='scikit-surgeryspeech',
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        description='Interface to speech services for image-guided surgery.',
        long_description=long_description,
        long_description_content_type='text/x-rst',
        url='https://github.com/SciKit-Surgery/scikit-surgeryspeech',
        author='Steve Thompson',
        author_email='s.thompson@ucl.ac.uk',
        license='BSD-3 license',
        classifiers=[
            'Development Status :: 3 - Alpha',

            'Intended Audience :: Developers',
            'Intended Audience :: Healthcare Industry',
            'Intended Audience :: Information Technology',
            'Intended Audience :: Science/Research',


            'License :: OSI Approved :: BSD License',


            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 3',

            'Topic :: Scientific/Engineering :: Information Analysis',
            'Topic :: Scientific/Engineering :: Medical Science Apps.',
        ],

        keywords='medical imaging',

        packages=find_packages(
            exclude=[
                'docs',
                'tests',
            ]
        ),


            
        install_requires=install_requires,

        entry_points={
            'console_scripts': [
                'sksurgeryspeech=sksurgeryspeech.__main__:main',
            ],
        },
    )
